
from pathlib import Path
TASK_REF_15NODES_LIST = [
    Path('module01.py'),
    Path('module02.py'),
    Path('module03.py'),
    Path('module04.py'),
    Path('module05.py'),
    Path('module06.py'),
    Path('module07.py'),
    Path('module08.py'),
    Path('module09.py'),
    Path('module10.py'),
    Path('module11.py'),
    Path('module12.py'),
    Path('module13.py'),
    Path('module14.py'),
    Path('module15.py'),
]


# EOF